package ut.maxorato.bsc.crawler.analysis;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.model.AnalysisContext;
import ut.maxorato.bsc.crawler.model.PathElement;
import ut.maxorato.bsc.crawler.model.PathElementType;
import ut.maxorato.bsc.crawler.model.UrlTemplate;
import ut.maxorato.bsc.crawler.utils.StringUtils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class InitialTemplateProducer implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(InitialTemplateProducer.class);

    private final AnalysisContext context;

    public InitialTemplateProducer(AnalysisContext context) {
        this.context = context;
    }

    private PathElement getNumOrTextValue(String text, boolean isGetPart) {
        if (StringUtils.isDigits(text)) {
            return new PathElement(PathElementType.NUM_VALUE, isGetPart, text);
        } else {
            return new PathElement(PathElementType.TEXT_STATIC, isGetPart, text);
        }
    }

    private boolean startsOrEndsWithNumber(String text) {
        return text.length() > 0 && (Character.isDigit(text.charAt(0)) || Character.isDigit(text.charAt(text.length() - 1)));
    }

    private List<PathElement> buildPathElements(String domain, String location, String parameters) {
        List<PathElement> elements = new ArrayList<>();
        elements.add(new PathElement(PathElementType.DOMAIN, false, domain));
        boolean mayHaveNumericParameter = false;

        if (location != null) {
            String[] parts = location.split("/");
            String lastPart = parts.length > 0 ? parts[parts.length - 1] : "";
            boolean addHtml = false;

            if (lastPart.endsWith(".html")) {
                parts[parts.length - 1] = lastPart.substring(0, lastPart.length() - 5);
            }

            for (String part : parts) {
                elements.add(new PathElement(PathElementType.SEPARATOR, false, "/"));
                elements.add(getNumOrTextValue(part, false));

                if (startsOrEndsWithNumber(part)) {
                    mayHaveNumericParameter = true;
                }
            }

            if (addHtml) {
                elements.add(new PathElement(PathElementType.SEPARATOR, false, ".html"));
            }
        }

        if (parameters != null) {
            String[] parts = parameters.split("&");
            boolean first = true;

            Arrays.sort(parts);

            for (String part : parts) {
                elements.add(new PathElement(PathElementType.SEPARATOR, true, first ? "?" : "&"));
                first = false;

                int equalPosition = part.indexOf("=");

                if (equalPosition == -1) {
                    elements.add(new PathElement(PathElementType.KEY, true, part));
                } else {
                    String key = part.substring(0, equalPosition);
                    String value = part.substring(equalPosition + 1);

                    elements.add(new PathElement(PathElementType.KEY, true, key));
                    elements.add(new PathElement(PathElementType.SEPARATOR, true, "="));
                    elements.add(getNumOrTextValue(value, true));

                    if (startsOrEndsWithNumber(value)) {
                        mayHaveNumericParameter = true;
                    }
                }
            }
        }

        if (!mayHaveNumericParameter) {
            return null;
        }

        return elements;
    }

    private String getDomainPart(String url, int firstSlashPosition, int parametersPosition, int hashPosition) {
        if (firstSlashPosition != -1) {
            return url.substring(0, firstSlashPosition);
        } else if (parametersPosition != -1) {
            return url.substring(0, parametersPosition);
        } else if (hashPosition != -1) {
            return url.substring(0, hashPosition);
        } else {
            return url;
        }
    }

    private String getLocationPart(String url, int firstSlashPosition, int parametersPosition, int hashPosition) {
        if (firstSlashPosition == -1) {
            return null;
        } else if (parametersPosition != -1) {
            return url.substring(firstSlashPosition + 1, parametersPosition);
        } else if (hashPosition != -1) {
            return url.substring(firstSlashPosition + 1, hashPosition);
        } else {
            return url.substring(firstSlashPosition + 1);
        }
    }

    private String getParametersPart(String url, int parametersPosition, int hashPosition) {
        if (parametersPosition == -1) {
            return null;
        } else if (hashPosition != -1) {
            return url.substring(parametersPosition + 1, hashPosition);
        } else {
            return url.substring(parametersPosition + 1);
        }
    }

    public List<PathElement> buildPathElements(String url) {
        int position = url.indexOf("://");

        if (position == -1) {
            throw new IllegalArgumentException("Invalid url, no schema");
        }

        position += 3;
        int firstSlashPosition = url.indexOf("/", position);
        int parametersPosition = url.indexOf("?", position);
        int hashPosition = url.indexOf("#", position);

        String domain = getDomainPart(url, firstSlashPosition, parametersPosition, hashPosition);
        String location = getLocationPart(url, firstSlashPosition, parametersPosition, hashPosition);
        String parameters = getParametersPart(url, parametersPosition, hashPosition);

        return buildPathElements(domain, location, parameters);
    }

    @Override
    public void run() {
        for (String url : context.urls) {
            List<PathElement> elements = buildPathElements(url);
            if (elements == null) {
                continue;
            }

            UrlTemplate template = new UrlTemplate(elements);
            UrlTemplate existing = context.initialTemplates.get(template.key);

            if (existing == null) {
                context.initialTemplates.put(template.key, template);
            } else {
                existing.merge(template);
            }
        }

        context.report.initialTemplateCount = context.initialTemplates.size();

        log.info("Produced {} initial templates.", context.initialTemplates.size());
    }
}
