package ut.maxorato.bsc.crawler.diff;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Attribute;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.nodes.Node;
import org.jsoup.nodes.TextNode;

import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.TreeMap;

public class DomDeltaCalculator {
    private boolean isIgnorableAttribute(String key) {
        if (key.startsWith("on") || "class".equals(key) || "style".equals(key) || "href".equals(key) || "action".equals(key) || "value".equals(key)) {
            return true;
        }

        return false;
    }

    private boolean compareNodeAttributes(Element base, Element other) {
        Map<String, String> attributes = new TreeMap<>();
        int hits = 0;

        for (Attribute attr : base.attributes()) {
            String lowerKey = attr.getKey().toLowerCase();

            if (!isIgnorableAttribute(lowerKey)) {
                attributes.put(lowerKey, attr.getValue());
            }
        }

        for (Attribute attr : other.attributes()) {
            String lowerKey = attr.getKey().toLowerCase();

            if (!isIgnorableAttribute(lowerKey)) {
                if (!attributes.containsKey(lowerKey) || !attributes.get(lowerKey).equals(attr.getValue())) {
                    return false;
                }

                hits++;
            }
        }

        return hits == attributes.size();
    }

    private boolean compareTexts(TextNode base, TextNode other) {
        String baseText = base.text();
        String otherText = other.text();

        if (baseText.equals(otherText)) {
            return true;
        } else if (baseText.contains("http://") || baseText.contains("https://")) {
            // Possible self-reference
            return true;
        }

        return false;
    }

    private boolean shallowEquals(Node base, Node other) {
        if (base.getClass() != other.getClass()) {
            return false;
        }

        String baseName = base.nodeName().toLowerCase();
        String otherName = other.nodeName().toLowerCase();

        if (!baseName.equals(otherName)) {
            return false;
        }

        return true;
    }

    private boolean compareElements(DomDelta delta, Stack<Integer> path, Element base, Element other) {
        String baseName = base.nodeName().toLowerCase();

        if (!shallowEquals(base, other)) {
            return false;
        }

        if (baseName.equals("title") || baseName.equals("meta")) {
            // Could use otherwise unused URL path elements
            return true;
        }

        if (!compareNodeAttributes(base, other)) {
            return false;
        }

        List<? extends Node> baseNodes = base.childNodes();
        List<? extends Node> otherNodes = other.childNodes();

        if (baseNodes.size() != otherNodes.size()) {
            return false;
        }

        DomDelta tempDelta = new DomDelta();

        for (int i = 0; i < baseNodes.size(); i++) {
            boolean result = true;
            Node baseNode = baseNodes.get(i);
            Node otherNode = otherNodes.get(i);

            path.push(i);

            if (!shallowEquals(baseNode, otherNode)) {
                return false;
            } else if (baseNode instanceof TextNode) {
                result = compareTexts((TextNode)baseNode, (TextNode)otherNode);
            } else if (baseNode instanceof Element) {
                result = compareElements(delta, path, (Element)baseNode, (Element)otherNode);
            }

            if (!result) {
                tempDelta.add(path, baseNode.outerHtml(), otherNode.outerHtml());
            }

            path.pop();
        }

        delta.differences.addAll(tempDelta.differences);

        return true;
    }

    public DomDelta calculate(String domOne, String domTwo) {
        Document base = Jsoup.parse(domOne);
        Document other = Jsoup.parse(domTwo);
        DomDelta delta = new DomDelta();
        Stack<Integer> path = new Stack<>();

        path.push(0);
        if (!compareElements(delta, path, base, other)) {
            delta.add(path, "", "");
        }
        path.pop();

        return delta;
    }
}
