package ut.maxorato.bsc.crawler.diff;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.model.RequestContext;

import java.util.Arrays;

public class PageComparisonStrategy {
    private static final Logger log = LoggerFactory.getLogger(PageComparisonStrategy.class);

    protected final HttpQueryExecutor httpQueryExecutor;
    protected final DomDeltaCalculator domDeltaCalculator;
    protected final String originalUrl;
    protected QueryResult originalResult;
    protected DomDelta originalVsSelf;

    public PageComparisonStrategy(RequestContext requestContext, String originalUrl) {
        this.httpQueryExecutor = requestContext.httpQueryExecutor;
        this.domDeltaCalculator = requestContext.domDeltaCalculator;
        this.originalUrl = originalUrl;
    }

    protected String normalizeUrl(String url) {
        int parameterLocation = url.indexOf('?');
        if (parameterLocation == -1) {
            return url;
        } else {
            String location = url.substring(0, parameterLocation);
            String[] parameters = url.substring(parameterLocation + 1).split("&");

            Arrays.sort(parameters);
            boolean first = true;

            for (String parameter : parameters) {
                location += (first ? '?' : '&') + parameter;
                first = false;
            }

            return location;
        }
    }

    protected String deslashUrl(String url) {
        while (url.length() > 0 && url.lastIndexOf('/') == url.length() - 1) {
            url = url.substring(0, url.length() - 1);
        }
        return url;
    }

    protected void loadOriginalResult() {
        if (originalResult == null) {
            originalResult = httpQueryExecutor.execute(originalUrl);

            if (!originalResult.isSuccess()) {
                log.debug("Expected working URL {} returned code {}", originalUrl, originalResult.code);
            } else {
                QueryResult repeatResult = httpQueryExecutor.execute(originalUrl);

                if (!repeatResult.isSuccess()) {
                    log.debug("Expected working URL {} returned code {}", originalUrl, originalResult.code);
                    originalResult = repeatResult;
                } else {
                    originalVsSelf = domDeltaCalculator.calculate(originalResult.text, repeatResult.text);
                }
            }
        }
    }

    protected DomDelta removePrefixes(DomDelta source, DomDelta filter) {
        DomDelta filtered = new DomDelta();

        for (DomDifference difference : source.differences) {
            boolean prefix = false;

            for (DomDifference exclude : filter.differences) {
                if (exclude.isPrefixOf(difference, false)) {
                    prefix = true;
                }
            }

            if (!prefix) {
                filtered.differences.add(difference);
            }
        }

        return filtered;
    }

    protected DomDelta filterEffectiveDelta(DomDelta source) {
        return removePrefixes(source, originalVsSelf);
    }

    protected void initialiseContextResult(ComparisonContext context) {
        if (context.alternateResult == null) {
            context.alternateResult = httpQueryExecutor.execute(context.alternateUrl);
        }
    }

    protected boolean isValidRedirectUrl(String url) {
        return originalUrl.equals(url);
    }

    protected Boolean responseCodeBasedComparison(ComparisonContext context) {
        loadOriginalResult();

        if (!originalResult.isSuccess()) {
            return false;
        }

        initialiseContextResult(context);
        if (context.alternateResult.isFailed()) {
            return false;
        } else if (context.alternateResult.isRedirect()) {
            String location = context.alternateResult.redirectLocation;

            return isValidRedirectUrl(location) ||
                    isValidRedirectUrl(normalizeUrl(location)) ||
                    (location.endsWith("/") && isValidRedirectUrl(deslashUrl(context.alternateResult.redirectLocation)));
        } else {
            return null;
        }
    }
}
