package ut.maxorato.bsc.crawler.generator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ut.maxorato.bsc.crawler.diff.GeneratedUrlComparisonStrategy;
import ut.maxorato.bsc.crawler.model.AnalysisContext;
import ut.maxorato.bsc.crawler.model.RequestContext;
import ut.maxorato.bsc.crawler.model.TemplateResult;
import ut.maxorato.bsc.crawler.model.UrlTemplate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class NewUrlGenerator implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(NewUrlGenerator.class);

    private static final int INITIAL_EDGE_X = 5;
    private static final int INITIAL_INTERVAL_Y = 10;
    private static final int MAXIMUM_EDGE_X = 40;
    private static final int MINIMUM_INTERVAL_Y = 1;
    private static final int FAILURE_THRESHOLD_MINIMUM = 40;
    private static final float FAILURE_THRESHOLD_RATIO = 0.05f;
    private static final int GLOBAL_THRESHOLD_MINIMUM = 500;
    private static final float GLOBAL_THRESHOLD_RATIO = 0.02f;
    private static final int MAX_TEMPLATE_QUERIES = 200;
    private static final int MAX_TEMPLATES_CHECKED = 10;

    private final AnalysisContext context;
    private final RequestContext requestContext;
    private final List<UrlTemplate> templates;
    private final List<TemplateResult> results;

    public NewUrlGenerator(AnalysisContext context, RequestContext requestContext) {
        this.context = context;
        this.templates = new ArrayList<>(context.finalTemplates.values());
        this.results = new ArrayList<>();
        this.requestContext = requestContext;
    }

    private void addGapIndices(Set<Integer> currentIterationIds, int start, int end, int edgeX, int intervalY) {
        if (end - start > 1000) {
            log.debug("Ignoring gap {} to {}, too large.", start, end);
            return;
        } else {
            log.trace("Adding indices from gap {} to {} with configuration {} {}", start, end, edgeX, intervalY);
        }

        if (start - end <= edgeX * 2) {
            for (int i = start; i < end; i++) {
                currentIterationIds.add(i);
            }
        } else {
            for (int i = start; i < start + edgeX; i++) {
                currentIterationIds.add(i);
            }

            for (int i = end - edgeX; i < end; i++) {
                currentIterationIds.add(i);
            }

            for (int i = start + edgeX; i < end - edgeX; i += intervalY) {
                currentIterationIds.add(i);
            }
        }
    }

    private boolean checkIndices(TemplateResult aggregateResult, TemplateResult templateResult, Set<Integer> indices, int checkedSize, GeneratedUrlComparisonStrategy strategy, String urlBase) {
        List<Integer> indexList = new ArrayList<>(indices);
        Collections.shuffle(indexList);

        int remaining = MAX_TEMPLATE_QUERIES - checkedSize;

        if (remaining == 0 && indices.size() > 0) {
            return false;
        }

        if (indices.size() > 0) {
            log.info("Checking a batch of {} entity indices for template {}", Math.min(remaining, indices.size()), urlBase);
        }

        for (Integer index : indexList) {
            String url = urlBase.replace("{N}", index.toString());
            boolean success = true;

            if (!strategy.isValidOtherEntityUrl(url)) {
                log.debug("Tested URL with failure: {}", url);
                success = false;
            } else {
                log.debug("Tested URL with success: {}", url);
            }

            aggregateResult.record(url, success);
            templateResult.record(url, success);

            if (aggregateResult.isFailure(GLOBAL_THRESHOLD_MINIMUM, GLOBAL_THRESHOLD_RATIO)) {
                log.info("Dropped the whole generation at template {}, global failure conditions reached.", urlBase);
                return false;
            } else if (templateResult.isFailure(FAILURE_THRESHOLD_MINIMUM, FAILURE_THRESHOLD_RATIO)) {
                log.info("Dropped genration for template {}, failure conditions reached.", urlBase);
                return false;
            }

            if (--remaining == 0) {
                return false;
            }
        }

        return true;
    }

    private TemplateResult processTemplate(TemplateResult aggregateResult, UrlTemplate template) {
        TemplateResult templateResult = new TemplateResult(template, true);
        Set<Integer> checkedIds = new HashSet<>();
        int edgeX = INITIAL_EDGE_X;
        int intervalY = INITIAL_INTERVAL_Y;

        String originalUrl = template.urlWithPlaceholder.replace("{N}", Integer.toString(template.knownValue));
        String emptyUrl = template.urlWithPlaceholder.replace("{N}", "999999999");

        GeneratedUrlComparisonStrategy strategy = new GeneratedUrlComparisonStrategy(requestContext, originalUrl, emptyUrl, template.getRegexPattern());

        while (intervalY >= MINIMUM_INTERVAL_Y && edgeX <= MAXIMUM_EDGE_X) {
            Set<Integer> currentIterationIds = new HashSet<>();
            int expectedValue = template.finalValues.get(0);

            for (int value : template.finalValues) {
                if (expectedValue != value) {
                    addGapIndices(currentIterationIds, expectedValue, value, edgeX, intervalY);
                }

                expectedValue = value + 1;
            }

            currentIterationIds.removeAll(checkedIds);

            if (!checkIndices(aggregateResult, templateResult, currentIterationIds, checkedIds.size(), strategy, template.urlWithPlaceholder)) {
                return templateResult;
            }

            checkedIds.addAll(currentIterationIds);

            intervalY /= 2;
            edgeX *= 2;
        }

        return templateResult;
    }

    @Override
    public void run() {
        TemplateResult aggregateResult = new TemplateResult(null, false);
        int remaining = MAX_TEMPLATES_CHECKED;

        templates.sort(new Comparator<UrlTemplate>() {
            @Override
            public int compare(UrlTemplate o1, UrlTemplate o2) {
                return o2.valueSetCount - o1.valueSetCount;
            }
        });

        context.report.allTemplates = templates;

        for (UrlTemplate template : templates) {
            TemplateResult result = processTemplate(aggregateResult, template);
            results.add(result);

            context.report.templateResults.add(result);

            log.info("Results for template {}: {} queries, {} success", template.urlWithPlaceholder, result.totalAttempts, result.generatedUrls.size());

            if (aggregateResult.isFailure(GLOBAL_THRESHOLD_MINIMUM, GLOBAL_THRESHOLD_RATIO)) {
                break;
            }

            if (--remaining == 0) {
                break;
            }
        }

        context.report.aggregateResults = aggregateResult;
    }
}
